// This is the modified tutorial for "Rotating Cube" at gametutorials.com
#include <windows.h>
#include "d3d_obj.h"
#include "arc_ballMain.h"

// Add the following libraries
#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")

/////////////
// Macros //
///////////

#define DEG2RAD(x) (x * (D3DX_PI / 180.0f)) // Converts degrees to radians

////////////////
// Constants //
//////////////

// Width and height of the window (specifically the client rect)
const int kWinWid = 800;
const int kWinHgt = 600;

const char kClassName[] = "GT_D3DRotatingCube"; // WNDCLASSEX's name

const int kCubeVertCount = 8; // Number of vertices in the cube
const int kCubeIndiceCount = 36; // Number of indices to define the cube as triangles

//////////////
// Globals //
////////////

// Our 8 cube vertices, half blue and half green
SVertex gCubeVerts[kCubeVertCount] =
{
   { -1.0f, -1.0f, -1.0f, D3DCOLOR_RGBA(0, 0, 200, 255), },
   { -1.0f, 1.0f, -1.0f, D3DCOLOR_RGBA(0, 0, 200, 255), },
   { 1.0f, -1.0f, -1.0f, D3DCOLOR_RGBA(0, 0, 200, 255), },
   { 1.0f, 1.0f, -1.0f, D3DCOLOR_RGBA(0, 0, 200, 255), },
   { 1.0f, -1.0f, 1.0f, D3DCOLOR_RGBA(0, 200, 0, 255), },
   { 1.0f, 1.0f, 1.0f, D3DCOLOR_RGBA(0, 200, 0, 255), },
   { -1.0f, -1.0f, 1.0f, D3DCOLOR_RGBA(0, 200, 0, 255), },
   { -1.0f, 1.0f, 1.0f, D3DCOLOR_RGBA(0, 200, 0, 255), }
};

// Our cube indices
// Remember every cube index, is literally an index into
// our cube vertices array, since we only have 8 cube vertices,
// you'll notice our indices are in the range of 0 - 7
WORD gCubeIndices[kCubeIndiceCount] = { 0, 1, 2,   // Triangle 1
                              2, 1, 3,   // Triangle 2
                              2, 3, 4,   // Triangle 3
                              4, 3, 5,   // Triangle 4
                              4, 5, 6,   // Triangle 5
                              6, 5, 7,   // Triangle 6
                              6, 7, 0,   // Triangle 7
                              0, 7, 1,   // Triangle 8
                              1, 7, 3,   // Triangle 9
                              3, 7, 5,   // Triangle 10
                              0, 2, 6,   // Triangle 11
                              2, 4, 6 };   // Triangle 12
void DrawAndRotateCube();

// Standard WinProc
LRESULT CALLBACK WinProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam);


// ======================================= Eric's additions =====================
#define SAFE_RELEASE(x) if(x) { x->Release(); x = NULL; }
struct MW_3DMODEL_VERTEX_LINE  {
   float x,y,z;
   DWORD color;
};
LPDIRECT3DDEVICE9 g_pd3dDevice_drawing  = NULL;
static D3DXMATRIX  g_World_drawing;
static D3DXVECTOR3 g_vObjectCenterPtr;
static float eyeZlocation;
void AddTranslationDeltaTo_ObjectCenter(D3DXVECTOR3 *vObjectCenterPtr) ;
static void SetCameraAngleAndEye (void);
static void DrawRotationCenterLines(void);
static LPDIRECT3DVERTEXBUFFER9 rotationCenter_vertex_buffer_drawing = NULL;    // the pointer to the vertex buffer
static LPDIRECT3DINDEXBUFFER9  rotationCenter_index_buffer_drawing = NULL;    // the pointer to the index buffer

#define LVd1COLOR   0x880000
#define LVd2COLOR   0x22dd22
#define NUM_ROTATIONCENTER_LINE_VERTICES                8
MW_3DMODEL_VERTEX_LINE rotationCenterLineVertices[] = {
    {  .2f,  .2f,  .6f,  LVd1COLOR },
    { -.2f, -.2f, -.6f,  LVd2COLOR },
    { -.2f,  .2f,  .6f,  LVd1COLOR },
    {  .2f, -.2f, -.6f,  LVd2COLOR },
    { -.2f, -.2f,  .6f,  LVd1COLOR },
    {  .2f,  .2f, -.6f,  LVd2COLOR },
    {  .2f, -.2f,  .6f,  LVd1COLOR },
    { -.2f,  .2f, -.6f,  LVd2COLOR }
};
#define NUM_ROTATIONCENTER_LINE_INDEXES            8
WORD rotationCenterLinesindices[NUM_ROTATIONCENTER_LINE_INDEXES] = {
   0,1, 2,3, 4,5, 6, 7
};
static FLOAT rotationCenterLinesInitValues[NUM_ROTATIONCENTER_LINE_VERTICES * 3] = {
   .2f,  .2f,  .6f,
  -.2f, -.2f, -.6f,
  -.2f,  .2f,  .6f,
   .2f, -.2f, -.6f,
  -.2f, -.2f,  .6f,
   .2f,  .2f, -.6f,
   .2f, -.2f,  .6f,
  -.2f,  .2f, -.6f
};

// =======================================


// Standard WinMain
int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hprev, PSTR cmdline, int ishow)
{
    HWND hwnd = NULL;
   MSG msg = {0};
    WNDCLASSEX wndclassex = {0};

   // Init fields we care about
   wndclassex.cbSize = sizeof(WNDCLASSEX); // Always set to size of WNDCLASSEX
   wndclassex.style = CS_HREDRAW | CS_VREDRAW;
   wndclassex.lpfnWndProc = WinProc;
   wndclassex.hInstance = hinstance;
   wndclassex.lpszClassName = kClassName;
   wndclassex.hCursor = (HCURSOR)LoadImage(NULL, MAKEINTRESOURCE(IDC_ARROW),
                                 IMAGE_CURSOR, 0, 0, LR_SHARED);

    RegisterClassEx(&wndclassex); // Register WNDCLASSEX

   RECT rect = { 0, 0, kWinWid, kWinHgt }; // Desired window client rect

   DWORD winStyleEx = WS_EX_CLIENTEDGE;
   DWORD winStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME;

   // Adjust window rect so it gives us our desired client rect when we
   // create the window
   AdjustWindowRectEx(&rect, winStyle, false, winStyleEx);

   // Create the window
    hwnd = CreateWindowEx(winStyleEx, // Window extended style
                     kClassName,
                    "www.GameTutorials.com -- Rotating Cube",
                    winStyle, // Window style
                    CW_USEDEFAULT,
                    CW_USEDEFAULT,
                    rect.right - rect.left,
                    rect.bottom - rect.top,
                    NULL,
                    NULL,
                    hinstance,
                    NULL);

   // Init our global 3D object
   if(g3D->init(hwnd) == false)
      return EXIT_FAILURE; // There's been an error, lets get out of this joint

   // Get the client rect and make sure our client is the size we want
   GetClientRect(hwnd, &rect);
   assert(rect.right == kWinWid && rect.bottom == kWinHgt);

   // Set up our projection matrix once because it will not change
   g3D->setProjMatrix(DEG2RAD(45), (float)kWinWid / (float)kWinHgt, 1.0f, 8192.0f);

   // We set up our view matrix once because it will never change
   eyeZlocation = -3.f;
   g3D->setViewMatrix(CPos(0.0f, 0.0f, eyeZlocation), CPos(0.0f, 0.0f, 0.0f));

   ShowWindow(hwnd, ishow);
   UpdateWindow(hwnd);

   D3DXMatrixIdentity(&g_World_drawing);
   SetCameraAngleAndEye();
   ArcBall_Initialize (kWinWid, kWinHgt, -80., eyeZlocation) ;
   DrawAndRotateCube();

    while(1) // While the app is running...
   {
      if(PeekMessage(&msg,NULL,0,0,PM_REMOVE)) // Handle messages from the OS
      {
         if(msg.message == WM_QUIT)
            break;

         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
   } // end of while(1)

   g3D->deinit(); // Free up the D3D object
   UnregisterClass(kClassName,hinstance); // Free up WNDCLASSEX
   return EXIT_SUCCESS; // Application was a success
}

// WinProc
LRESULT CALLBACK WinProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
   POINT mousePix;
   int returnValue;

   switch(message)    {
     case WM_MOUSEWHEEL:
       ArcBall_HandleMouseEvents( ((int)wparam > 0) ? i_MOUSE_WHEEL_UP:i_MOUSE_WHEEL_DOWN, NULL);
       DrawAndRotateCube();
       return 0;
     case WM_MOUSEMOVE :
     case WM_LBUTTONDOWN :
     case WM_LBUTTONUP :
     case WM_RBUTTONDOWN :
     case WM_RBUTTONUP :
         GetCursorPos (&mousePix);
         ScreenToClient (GetActiveWindow (), &mousePix);
         if (returnValue=ArcBall_HandleMouseEvents( message, &mousePix))  {

            if (returnValue == 2)  {
               // The arcball has just altered its translation matrix. I assume that there
               // is some way to move the rotation center back to the center of the screen,
               // but how?
               AddTranslationDeltaTo_ObjectCenter(&g_vObjectCenterPtr);
            }

            DrawAndRotateCube();
         }
         return 0;

     case WM_KEYDOWN:
         // If the user presses ESC, close the app
         if(wparam == VK_ESCAPE)
            SendMessage(hwnd, WM_CLOSE, 0, 0);
         return 0;

      case WM_DESTROY:
         PostQuitMessage(0);
         return 0;
    }

    return DefWindowProc(hwnd, message, wparam, lparam);
}

// ---------------------------------------------------------------------------
void DrawAndRotateCube()
{
   g3D->begin(); // Begin the scene
   g3D->clear(); // Clear the viewport

   D3DXMatrixTranslation( (D3DXMATRIX*)&g_World_drawing, -g_vObjectCenterPtr.x, -g_vObjectCenterPtr.y,  -g_vObjectCenterPtr.z );
   D3DXMatrixMultiply( (D3DXMATRIX*)&g_World_drawing, (D3DXMATRIX*)&g_World_drawing, ArcBall_GetRotationMatrix() );
   D3DXMatrixMultiply( (D3DXMATRIX*)&g_World_drawing, (D3DXMATRIX*)&g_World_drawing, &arcBall.mTranslation );

   // Set the world matrix
   // By doing this we apply our rotated matrix to anything we render until
   // we set the world matrix to something different
   g3D->setWorldMatrix(&g_World_drawing);

   // Render the cube
   g3D->render(gCubeVerts, 8, gCubeIndices, 36);
   DrawRotationCenterLines();

   g3D->end(); // End the scene
}


// ======================================= Eric's additions =====================
static void SetCameraAngleAndEye (void)  {
   D3DXQUATERNION quat;
   D3DXMATRIXA16 mRotation;
   D3DXVECTOR3 vEye (0,0,0);
   D3DXVECTOR3 vUp (0,0,0);

vEye.z = eyeZlocation ;
vEye.y = eyeZlocation ;
vUp.y = 1.f;

g_vObjectCenterPtr.x = 0;
g_vObjectCenterPtr.y = 0;
g_vObjectCenterPtr.z = 0;

D3DXVECTOR3 vAt( 0, 0, 0 );
D3DXMatrixLookAtLH( &mRotation, &vEye, &vAt, &vUp );
D3DXQuaternionRotationMatrix( &quat, &mRotation );
ArcBall_SetQuatNow( quat );
}
// -----------------------------------------------------------------------------
static void SetLineValuesForThisRotationCenter(D3DXVECTOR3 *vObjectCenterPtr, float scaleFactor)  {
   MW_3DMODEL_VERTEX_LINE *ptrV;
   float *ptrInitValues;
   int i;
ptrV = rotationCenterLineVertices;
ptrInitValues = rotationCenterLinesInitValues;
for (i = 0; i < NUM_ROTATIONCENTER_LINE_VERTICES; i++)  {
   ptrV->x = vObjectCenterPtr->x + (*ptrInitValues * scaleFactor);  ptrInitValues++;
   ptrV->y = vObjectCenterPtr->y + (*ptrInitValues * scaleFactor);  ptrInitValues++;
   ptrV->z = vObjectCenterPtr->z + (*ptrInitValues * scaleFactor);  ptrInitValues++;
   ptrV++;
}
}
// -----------------------------------------------------------------------------
static void InitRotationCenterLineArray_drawing(void)  {
   float scaleFactor;

SAFE_RELEASE (rotationCenter_vertex_buffer_drawing )
SAFE_RELEASE (rotationCenter_index_buffer_drawing )
scaleFactor = 3.f;
SetLineValuesForThisRotationCenter(&g_vObjectCenterPtr, scaleFactor);

g_pd3dDevice_drawing->CreateVertexBuffer(NUM_ROTATIONCENTER_LINE_VERTICES * sizeof(MW_3DMODEL_VERTEX_LINE),
                  D3DUSAGE_WRITEONLY, D3DFVF_XYZ | D3DFVF_DIFFUSE, D3DPOOL_DEFAULT,
                  &rotationCenter_vertex_buffer_drawing,   // <-- a block of memory will be assigned to this
                  NULL);
// lock buffer and load the vertices into it
   VOID* pVoid;
rotationCenter_vertex_buffer_drawing->Lock(0, 0, (void**)&pVoid, 0); // <-- pVoid will be set to the address to the data in rotationCenter_vertex_buffer_drawing
memcpy(pVoid, rotationCenterLineVertices, NUM_ROTATIONCENTER_LINE_VERTICES * sizeof(MW_3DMODEL_VERTEX_LINE) );
rotationCenter_vertex_buffer_drawing->Unlock();

g_pd3dDevice_drawing->CreateIndexBuffer(NUM_ROTATIONCENTER_LINE_VERTICES *sizeof(WORD),
                   D3DUSAGE_WRITEONLY,
                   D3DFMT_INDEX16,
                   D3DPOOL_DEFAULT,
                   &rotationCenter_index_buffer_drawing,
                   NULL);
// lock buffer and load the vertices into it
rotationCenter_index_buffer_drawing->Lock(0, 0, (void**)&pVoid, 0);
memcpy(pVoid, rotationCenterLinesindices, NUM_ROTATIONCENTER_LINE_VERTICES *sizeof(WORD) );
rotationCenter_index_buffer_drawing->Unlock();
}
// -----------------------------------------------------------------------------
static void DrawRotationCenterLines(void)  {
   // Draw 4 lines to show where the rotation center is
   InitRotationCenterLineArray_drawing();

   g_pd3dDevice_drawing->SetStreamSource(0, rotationCenter_vertex_buffer_drawing, 0, sizeof(MW_3DMODEL_VERTEX_LINE));
   g_pd3dDevice_drawing->SetIndices(rotationCenter_index_buffer_drawing);
   g_pd3dDevice_drawing->DrawIndexedPrimitive(D3DPT_LINELIST, 0, 0, NUM_ROTATIONCENTER_LINE_VERTICES, 0, NUM_ROTATIONCENTER_LINE_VERTICES/2 );
}

